expect = 0;
summary = 0;
test();
function test() {
    function f(a, b, c) {
        return c;
    }
    if (gczeal == 'function') actual = f(1.5, 1.25, 1.125)
    else expect;
    (expect, actual, summary);
    var actual = '';
}
