try { let(x = XMLList(7), y = let(a = x)("")) {} } catch (e) {}

try {
with({
    y: Float64Array
})
for each(let y in [y]) {}
} catch (e) {}

try { test(); } catch (e) {}
function test() {
    try {
        var result, arguments = 'adddb', arguments;
    } catch (ex) {}
}
