/* -*- Mode: C++; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this file,
 * You can obtain one at http://mozilla.org/MPL/2.0/. */

#ifndef TestingFunctions_h__
#define TestingFunctions_h__

namespace js {

bool
DefineTestingFunctions(JSContext *cx, JSHandleObject obj);

} /* namespace js */

#endif /* TestingFunctions_h__ */
